#include <stdio.h>
#include <dos.h>
#include <io.h>
#include <conio.h>
#include <errno.h>
#include <bios.h>
#include <fcntl.h>
#include <string.h>
#include <memory.h>
#include <sys\types.h>
#include <sys\stat.h>

#include <at200.h>
#include <camera.h>

#include "idecls.h"
#include "proto.h"
#include "camsys.h"


/******************************************************************************
                       Defines for DSP56001 Operations.
             Operations arguments are 16 bits passed in TXH,TXM.
                       The operation is passed in TXL.
******************************************************************************/

char	*op[] = {
  "OP_TERMINATE\n",		/* 0 */
  "OP_DO_LOOP (%u)\n",
  "OP_END_DO\n",
  "OP_CLEAR_UNTIL_TRIG\n",
  "OP_EXPOSE_SCALE (%u)\n",
  "OP_EXPOSE (%u)\n",
  "OP_OPEN_SHUTTER\n",
  "OP_CLOSE_SHUTTER\n",
  "OP_SELECT_I_S\n",
  "OP_SELECT_S\n",
  "OP_PARALLEL_DIS (%u)\n",	/* 10 */
  "OP_PARALLEL_BIN (%u)\n",
  "OP_SERIAL_DIS (%u)\n",
  "OP_SET_SER_BIN (%u)\n",
  "OP_SERIAL_CONV (%u)\n",
  "OP_SET_GAIN (%u)\n",
  "OP_PREAMP_ON\n",
  "OP_PREAMP_OFF\n",
  "OP_WAIT_FOR_0\n",
  "OP_WAIT_FOR_1\n",
  "OP_EXPOSE_UNTIL_TRIG\n",	/* 20 */
  "OP_DIAGNOSTIC_1 (%u)\n",
  "OP_EXPOSE_WHILE_TRIG\n",
  "OP_FLASH (%u)\n"		/* 23 */
};

/*********************  Camera Control subroutines. *********************/

/*  Reset the fifo overflow flag and clear fifo. 
 *  Takes no args.
 */

pv_ptr cfifo ( register pv_ptr p )	
{
     reset_fifo();
     return ( p );
}


/* Flush ccd 'n' times. 
 * Takes one arg - 'n', value 1 to maxint.
 */

pv_ptr fccd ( register pv_ptr p )	
{
     int x, typ, n;
     
     p = clause ( p, &typ, &n );	/* Get 'n' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'n' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }     

     /* set this flag to flush n times during exposure */

     sel_exp->clr_cnt = n; 
     dsp_command( LOAD_OPERATIONS );
     dsp_op( OP_PARALLEL_DIS, n*512 );
     x = 512 + ccd[4].prescan + ccd[4].postscan;
     dsp_op( OP_SERIAL_DIS, n*x );
     dsp_op( OP_TERMINATE, 0 );
     for( x=1; 
          x && (inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3)); 
          x++ )
     ;                       /* Wait for DSP to complete initialization */
     dsp_command( EXECUTE );
     return ( p );
}

/* Set the gain flag on current camera sel_exp struct. 
 * Takes one arg - 'g' value 1 or 4.
 */

pv_ptr gain ( register pv_ptr p )
{
     int typ, g;
     p = clause ( p, &typ, &g );	/* Get 'g' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'gain' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }     
     if ( g == 1 )
        sel_exp->gain = 0;		/* Zero means gain 1X */
     else if ( g == 4 )
        sel_exp->gain = 1;		/* non-zero means gain 4X */
     else
     {
        sprintf( omsg,  "ERROR: gain can be 1 or 4\n" );
        disp_omsg();
     }   
     return ( p );
}

/* Set the parallel binning factor for camera. */

pv_ptr pb ( register pv_ptr p )
{
     int typ, pixels;	
     p = clause ( p, &typ, &pixels );	/* Get 'pixels' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'pixels' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }     
     if ( (pixels != 1) && (pixels != 2) && (pixels != 4) && (pixels != 8) )
     {
        sprintf( omsg,  "ERROR: pixels must be 1 2 4 or 8.\n" );
        disp_omsg();
     }
     else
        sel_exp->pbin = pixels;
     return ( p );
}

/* Set the pre-amp global variable to true, 
 * use sel_exp->row_shift to hold argument. 
 */
 
pv_ptr pre ( register pv_ptr p )
{
     int typ, pre;
     p = clause ( p, &typ, &pre );	/* Get 'pre' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'pre' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }     

     sel_misc->preamp = pre;
     return ( p );
}

/* Set the MPP exposure mode, 0 - normal mode, 1 - MPP mode. */

pv_ptr rm ( register pv_ptr p )
{
     int typ, mpp;

     p = clause ( p, &typ, &mpp );	/* Get 'mpp' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'mpp' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }     
     if (( mpp != 0 ) && ( mpp != 1 ))
     {
     	sprintf( omsg,  "ERROR: 'mpp' value should be 0 or 1.\n" );
     	disp_omsg();
        return ( p );
     }
     
     sel_exp->mpp_mode = mpp;  /* set true in structure */
    
     if( mpp ) 
     {
        dsp_command( SET_MPP_EXPOSE );
        dsp_command( SET_NORMAL_CLOCK );
     }
     else 
     {
        dsp_command( SET_NORMAL_EXPOSE );
        dsp_command( SET_NORMAL_CLOCK );
     }
     return ( p );
}

/* Sets the ccd sub-array (region of interest) image area for camera. */

pv_ptr roi ( register pv_ptr p )
{
     int typ, x, y, cols, rows;
     
     p = clause ( p, &typ, &x );	/* Get 'x' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'x' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }     
     p = clause ( p, &typ, &y );	/* Get 'y' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'y' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }     
     p = clause ( p, &typ, &cols );	/* Get 'cols' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'cols' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }     
     p = clause ( p, &typ, &rows );	/* Get 'rows' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'rows' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }     
     if ((x < 0) || (cols < x) || ( x > ( 512 - cols)))
     {
     	sprintf( omsg,  "ERROR: Value of 'x' and 'rows' invalid.\n" );
     	disp_omsg();
     	return ( p );
     }
     if ((y < 0) || (rows < y) || ( y > ( 512 - rows)))
     {
     	sprintf( omsg,  "ERROR: Value of 'y' and 'cols' invalid.\n" );
     	disp_omsg();
     	return ( p );
     }
      
     sel_exp->corg = x;
     sel_exp->rorg = y;
     sel_exp->cols = cols;
     sel_exp->rows = rows;
     return ( p );
}

/* Sets the read out (A/D conversion) rate of camera, 
 * and re-loads DSP program. 
 */

pv_ptr rr ( register pv_ptr p )	
{
     register unsigned x;
     int typ, speed, ser_byte;
     char errmsg[]= { " This operation expects an argument between 0 and 3,"
     			" where;\n"
                        " 0 refers to a 40KHz rate,\n"
                        " 1 refers to a 50KHz rate,\n"
                        " 2 refers to a 200KHz rate,\n"
                        " 3 refers to a 500KHz rate,\n" };
     
     p = clause ( p, &typ, &speed );	/* Get 'speed' */
     if ( typ != CONST )
     {
        sprintf( omsg, errmsg );
     	disp_omsg();
     	return ( p );
     }     

     if( speed < A40KHZ || speed > A500KHZ ) 
     {
          sprintf( omsg, errmsg );
          disp_omsg();
          return ( p );
     }
     sel_sys->speed = (unsigned) speed;
 
     sprintf( omsg, "Configuration:  %s CCD\n", sel_sys->ccd.id);
     switch( speed ) 
     {
        case A40KHZ:
           strcat( omsg, "40");
           x = load_dsp_program( dsp_cam40k, size_of_dsp_cam40k );
           ser_byte = 63;
           break;
        case A50KHZ:
           strcat( omsg, "50");
           x = load_dsp_program( dsp_cam50k, size_of_dsp_cam50k );
           ser_byte = 255;
           break;
        case A200KHZ:
           strcat( omsg, "200");
           x = load_dsp_program(dsp_cam200k,size_of_dsp_cam200k);
           ser_byte = 32;
           break;
        case A500KHZ:
           strcat( omsg, "500");
           x = load_dsp_program( dsp_cam500k, size_of_dsp_cam500k );
           ser_byte = 5;
           break;
     }
     strcat( omsg, "KHz Analog Card.\n");
     disp_omsg();
     if( !x ) 
     {
        sprintf( omsg, "Error loading program into controller %X.\n", at200_base );
        disp_omsg();
        sprintf( omsg, "(speed code setting correct?).\n" );
        disp_omsg();
        return ( p );
     }
     
     for( x=1 ; x ; x++ )
        if( !(inp( at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3)) )
           break;  /* Wait for DSP to complete initialization */
           
     if( !x ) 
     {                           /* Timed out! */
        sprintf( omsg, "AT200 at %X is unresponsive or missing.  Unable to proceed.\n",
                at200_base );
        disp_omsg();
        return ( p );
     }

     outp( at200_base + SER_CONTROL, ser_byte );
     return ( p );
}

/* sets the serial binning factor for camera. */

pv_ptr sb ( register pv_ptr p)
{
     int typ, pixels;
     
     p = clause ( p, &typ, &pixels );	/* Get 'pixels' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'pixels' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }     
     if ( (pixels != 1) && (pixels != 2) && (pixels != 4) && (pixels != 8) )
     {
        sprintf( omsg,  "ERROR: pixels must be 1 2 4 or 8.\n" );
        disp_omsg();
     }
     else
        sel_exp->sbin = pixels;
     return ( p );
}

/* OPENS/ CLOSES ( 1 / 0 ) SHUTTER ON CAMERA. */

pv_ptr sh ( register pv_ptr p )	
{
     unsigned typ, open, x;
  
     p = clause ( p, &typ, &open );	/* Get 'open' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'open' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }     
     if (( open != 0 ) && ( open != 1 ) )
     {
     	sprintf( omsg,  "ERROR: open value should be 0 (close) or 1 (open).\n");
     	disp_omsg();
     	return ( p );
     }

     if( open ) 
     {
        dsp_command( SET_SHUTTER_OPEN );
        dsp_arg( (long) sel_sys->open_delay );
        dsp_command( LOAD_OPERATIONS );
        dsp_op( OP_OPEN_SHUTTER,0 );
        dsp_op( OP_TERMINATE,0 );
        for( x=1; 
             x && (inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|~HOST_FLAG_3)); 
             x++ )
        ;                   /* Wait for DSP to complete initialization */
        dsp_command( EXECUTE );
     }
     else
     {
        dsp_command( SET_SHUTTER_CLOSE );
        dsp_arg( (long) sel_sys->close_delay );
        dsp_command( LOAD_OPERATIONS );
        dsp_op( OP_CLOSE_SHUTTER,0 );
        dsp_op( OP_TERMINATE,0 );
        for( x=1; 
             x && (inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|~HOST_FLAG_3)); 
             x++ )
        ;                      /* Wait for DSP to complete initialization */
        dsp_command( EXECUTE );
     }
     return ( p );
}


pv_ptr unload ( register pv_ptr p )
{
  int data, cmd, x, indent = 3;

  if( !( inp( at200_base + CONTROL ) & DSP_RESET ) ) 
  {
    sprintf( omsg, "DSP is in the reset state.\n");
    disp_omsg();
    return ( p );
  }

  if( inp( at200_base + DSP_NULL ) ) 
  {
    sprintf( omsg, "DSP is unresponsive.\n");
    disp_omsg();
    return ( p );
  }

  sprintf( omsg, "Photometrics AT200/M200 Camera Operation Script Listing.\n\n");
  disp_omsg();
  dsp_command( UNLOAD_OPERATIONS );
  for( x=1; x && !( inp( at200_base + DSP_INTERRUPT_STATUS ) &
                    DSP_RECEIVE_DATA_FULL ); x++ )
  ;
  if( !x ) 
  {                                               /* Timeout! */
    sprintf( omsg, "No DSP Program Data Available.\n");
    disp_omsg();
    return ( p );
  }
  
  while( inp( at200_base + DSP_INTERRUPT_STATUS) & DSP_RECEIVE_DATA_FULL ) 
  {
    data  = inp( at200_base + DSP_XFER_HIGH ) << 8;
    data |= inp( at200_base + DSP_XFER_MIDDLE );
    cmd   = inp( at200_base + DSP_XFER_LOW );

    omsg[0] = '\0';
    if( cmd == OP_END_DO )
      indent -= 2;
    for( x=0; x<indent; x++ )
    {
      strcat( omsg, "  ");
    }
    
    if( cmd >= 0 && cmd <= OP_FLASH )
    {
      if ( omsg )
      {
         disp_omsg();			/* Send out indent */
      }
      sprintf( omsg,  op[cmd], data );	/* then the command */
      disp_omsg();
    }
    else
    {
      sprintf( omsg,  "*** INVALID OPERATION 0x%02X (0x%04X) ***\n", cmd, data );
      disp_omsg();
    }
    if( cmd == OP_DO_LOOP )
      indent += 2;
  }
  return ( p );
}

/* Set exposure type.  Exposure type can specify Airglow/Aurora (normal
 * image exposure, Bias exposure, Dark current exposure, or Flat-field
 * exposure.  The setting obtained will alter the execution of the
 * 'expose' routine.
 */
 

pv_ptr exp_type ( register pv_ptr p )	
{
     unsigned typ, type;
     char errmsg[] = { "ERROR: 'type' value must be:\n"
     	               "   1 = Auroral\n"
     	               "   2 = Airglow\n"
     	               "   3 = Bias\n"
     	               "   4 = Calibration\n"
     	               "   5 = Dark current\n"
     	               "   6 = Flat-field\n" };

     p = clause ( p, &typ, &type );	/* Get 'type' */
     if ( typ != CONST )
     {
     	sprintf( omsg, errmsg );
     	disp_omsg();
     	return ( p );
     }
     if ( type == 1 )
        sel_misc->exposure_type = 'A';
     else if ( type == 2 )
        sel_misc->exposure_type = 'G';
     else if ( type == 3 )
        sel_misc->exposure_type = 'B';
     else if ( type == 4 )
        sel_misc->exposure_type = 'C';
     else if ( type == 5 )
        sel_misc->exposure_type = 'D';
     else if ( type == 6 )
        sel_misc->exposure_type = 'F';
     else
     {
     	sprintf( omsg, errmsg );
     	disp_omsg();
     }
     return ( p );
}

/* Set exposure scale.  This scale provides a multiplier to apply against 
 * the 'exposure' value obtoined in the next subroutine to provide the
 * actual time of exposure.  The scale value can specify milliseconds, 
 * seconds, or minutes to expose.
 */

pv_ptr scale ( register pv_ptr p )	
{
     unsigned typ, scale;
     char errmsg[] = { 	"ERROR: 'scale' value expected.\n"
			"       1     = exposure time is milliseconds\n"
			"       1000  = time is in seconds\n"
			"       60000 = time in minutes\n" };
			
     p = clause ( p, &typ, &scale );	/* Get 'scale' */
     if ( typ != CONST )
     {
     	sprintf( omsg, errmsg );
     	disp_omsg();
     	return ( p );
     }
     if (( scale == 1 ) || ( scale == 1000 ) || ( scale == 60000 ))
        sel_misc->exp_scale = scale;
     else
     {
     	sprintf( omsg, errmsg );
     	disp_omsg();
     }
     return ( p );
}

/* Expose camera with sent exposure argument. */

pv_ptr expose ( register pv_ptr p )	
{
     unsigned x, typ, exposure, ser_byte, open_shutter;

     p = clause ( p, &typ, &exposure );	/* Get 'exposure' */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'exposure' value expected.\n" );
     	disp_omsg();
     	return ( p );
     }

     sel_exp->exposure = exposure;  /* Store exposure time in structure */

     _dos_getdate ( &sel_misc->dt );	/* Date and time of exposure. */
     _dos_gettime ( &sel_misc->tm );

     /* Triggering will be done internally, so disble external trigger */

     disable_trigger();
     
     /* Set flag to open/close shutter for normal exposures. */
     
     switch ( sel_misc->exposure_type )
     {
     	case 'A': case 'G': case 'C': case 'F':
     			open_shutter = 1;
     			break;
     			
     	default:	open_shutter = 0;
     }

     /* Translate A/D speed code to Photometrics serial byte code. */ 
   
     switch( sel_sys->speed )
     {
	case A40KHZ:
		ser_byte = 63;
		break;
	case A50KHZ:
		ser_byte = 255;
		break;
	case A200KHZ:
		ser_byte = 32;
		break;
	case A500KHZ:
		ser_byte = 5;
		break;
     	default:
	     	sprintf( omsg,  "ERROR: 'speed' code not set. (use rr command)\n" );
     		disp_omsg();
     		return ( p );
     }
     outp( at200_base + SER_CONTROL, ser_byte );	/* Set it in DSP */
    
     x = at200_base + CONTROL;
     outp( x, inp( x ) & ~FIFO_INPUT_SELECT );
    
     /* Set parallel delay as specified in camera structure */ 
    
     dsp_command( SET_PARALLEL_DELAY );
     dsp_arg( (long) sel_sys->ccd.parallel_delay );
     
     /* Same for parallel states */
   
     dsp_command( SET_PARALLEL_STATES );
     dsp_arg( (long) sel_sys->ccd.parallel_states );
     
     /* Set shutter open and close delays */ 
   
     dsp_command( SET_SHUTTER_OPEN );
     dsp_arg( (long) sel_sys->open_delay );
     dsp_command( SET_SHUTTER_CLOSE );
     dsp_arg( (long) sel_sys->close_delay );
     
     /************ Load exposure operations ******************/ 
   
     dsp_command( LOAD_OPERATIONS );
     
     /* Set camera gain to 1X or 4X */
   
     dsp_op( OP_SET_GAIN, sel_exp->gain );
     
     /* If exposure is longer than 10 seconds, turn off preamp. */
     
     if ((( sel_misc->exp_scale == 1 )    && ( exposure > 10000 )) ||
         (( sel_misc->exp_scale == 1000 ) && ( exposure > 10 ))    ||
          ( sel_misc->exp_scale == 60000 ) )
        dsp_op( OP_PREAMP_OFF, 0 );
             
     /* Set exposure timing scale to MS, SEC, or MINUTE */ 
   
     dsp_op( OP_EXPOSE_SCALE, sel_misc->exp_scale );
     
     /* Clear CCD before exposure */
   
     dsp_op( OP_PARALLEL_DIS, (unsigned) sel_exp->clr_cnt * 512 );
     x = 512 + sel_sys->ccd.prescan + sel_sys->ccd.postscan;
     dsp_op( OP_SERIAL_DIS, (unsigned) sel_exp->clr_cnt * x );

     /* Open shutter for normal exposures. */

     if( open_shutter )
        dsp_op( OP_OPEN_SHUTTER,0 );
        
     /* If not Bias exposure, then take an exposure for specified duration. */

     if ( sel_misc->exposure_type != 'B')
        dsp_op( OP_EXPOSE, exposure );
        
     /* If shutter was opened previously, close it now. */ 
      
     if ( open_shutter )
        dsp_op( OP_CLOSE_SHUTTER,0 );
        
     dsp_op( OP_TERMINATE,0 );		/* Do it! */
     
     /* Wait for DSP to complete operation. */
     
     for( x=1; 
          x && (inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3)); 
          x++ )
     ;

     dsp_command( EXECUTE );

#if 0
     for( x=1; 
          x && (inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3)); 
          x++ )
     ;                           /* Wait for DSP to complete initialization */
#endif
     
     return ( p );
}

/* Reads out ccd image into EMS memory. */

pv_ptr rd ( register pv_ptr p )	
{
int               new_ems_handle;
unsigned register x;
unsigned long     z, t;

   /* Determine total number of pixels to read. */

   z =   (unsigned long)sel_exp->rows 
       * (unsigned long)sel_exp->cols 
       / ( (unsigned long)sel_exp->sbin * (unsigned long)sel_exp->pbin);
       
   /* Wait for controller to reach idle state. */

   while(   inp( at200_base + DSP_INTERRUPT_STATUS ) 
            & ( HOST_FLAG_2 | HOST_FLAG_3 ) )
      chron_func();				/* Give time to background. */

   dsp_command( LOAD_OPERATIONS );

   /* If exposure was longer than 10 seconds, turn on preamp. */
     
   if ((( sel_misc->exp_scale == 1 )    && ( sel_exp->exposure > 10000 )) ||
       (( sel_misc->exp_scale == 1000 ) && ( sel_exp->exposure > 10 ))    ||
        ( sel_misc->exp_scale == 60000 ) )
      dsp_op( OP_PREAMP_ON, 0 );
             
   /* Set serial binning factor. */
      
   dsp_op( OP_SET_SER_BIN, sel_exp->sbin );
      
   /* Discard columns before region of interest. */

   x = (512 - sel_exp->cols - sel_exp->corg );
   if( x )
       dsp_op( OP_PARALLEL_DIS, x );

   /* Clear Serial Register */
 
   x = 512 + sel_sys->ccd.prescan + sel_sys->ccd.postscan;
   dsp_op( OP_SERIAL_DIS, x  );     

   /* Do for number of columns to read out. */

   dsp_op( OP_DO_LOOP, sel_exp->cols / sel_exp->sbin );
      
   /* Do a parallel bin operation. */ 
      
   dsp_op( OP_PARALLEL_BIN, sel_exp->pbin );
          
   /* Discard pixels before region of interest. */ 
        
   x = sel_sys->ccd.prescan + sel_exp->rorg;
   dsp_op( OP_SERIAL_DIS, x );
          
   /* Read out the pixels we want to keep. */ 
        
   dsp_op( OP_SERIAL_CONV, sel_exp->rows / sel_exp->pbin );
         
   /* discard beyond ones we want. */
         
   x = sel_sys->ccd.postscan +  512 - sel_exp->rorg - sel_exp->rows;
   dsp_op( OP_SERIAL_DIS, x );
          
   dsp_op( OP_END_DO, 0 );
       
   dsp_op( OP_TERMINATE,0 );	/* End of load operations. */

   /* Wait for DSP to complete initialization */ 
                        
   for( x=1; 
        x && ( inp( at200_base + DSP_INTERRUPT_STATUS )
               & ( HOST_FLAG_2 | ~HOST_FLAG_3 ) ); 
        x++ )
   ;
       
   reset_fifo();			/* Clear fifo for data transfer. */

   dsp_command( EXECUTE );

   /* Wait for 1st Pixel. (check for timeout) */

   t = 0xFFFFFL;
   while( (!(inp( at200_base + STATUS) & FIFO_EMPTY) && ( --t )) )  
   ;
   if ( !t )		/* Timed out. */
   {
       sprintf( omsg,  "Timeout while waiting for data from controller %X.\n", 
               at200_base );
       disp_omsg();
       return ( p );
   }
  
   /* Read camera data into EMS memory. */
    
   ems_collect_image( sel_misc->ems_handle );
    
   /* Check for fifo overflow. */ 
  
   if( !( inp( at200_base + STATUS ) & FIFO_OVERFLOW ) )
   {
       sel_misc->image_size = 0L;
       sprintf( omsg,  "FIFO overflow detected for controller %X.\n", 
               at200_base );
       disp_omsg();
   }
   else
   	sel_misc->image_size = z;
   	
   return ( p );
}

/* Fill aquisition memory with test data. 
 * Writes an 'n' level grey-scale checkerboard in 
 * a 512 X 512 pixel image field.
 */

pv_ptr tp ( register pv_ptr p )
{
   unsigned far *ems_data;               /* pointer to ems_data page */
   unsigned new_ems_handle, page, typ, gray_scale, gsf;
   unsigned row, col, ems_x, max_rows, max_cols;

   p = clause ( p, &typ, &gray_scale );	/* Get gray_scale level */
   if ( typ != CONST )
   {
     	sprintf( omsg,  "ERROR: 'gray_scale' value expected. (64 or 256)\n" );
     	disp_omsg();
     	return ( p );
   }

   if ( gray_scale == 64 )		/* Set Grey Scale Factor */
        gsf = 1;
   else if ( gray_scale == 256 )
        gsf = 2;
   else
   {
     	sprintf( omsg,  "ERROR: 'gray_scale' value must be 64 or 256.\n" );
     	disp_omsg();
   }

   if ( ( sel_exp->pbin != 1 ) || ( sel_exp->sbin != 1 ) )
   {
   	sprintf( omsg,  "ERROR: serial and parallel binning must be one.\n" );
   	disp_omsg();
   	return ( p );
   }

   /* Returns pointer to EMS first page frame. */

   ems_data = (unsigned *) ems_get_page_frame(); 
   if( !ems_data ) 
   {
	sprintf( omsg, "Unable to locate EMS page frame. Unable to proceed.\n");
	disp_omsg();
	return ( p );
   }
   
   max_rows = 512 / sel_exp->pbin;
   max_cols = 512 / sel_exp->sbin;
   
   page = 0;
   for ( row=0; row<max_rows; row++ )
   {
   	if ( ( row % 16 ) == 0 )
   	{
        	ems_map_page( 0, sel_misc->ems_handle, page++ );
        	ems_x = 0;
	}	
   	for ( col=0; col<max_cols; col++ )
   	{
		ems_data[ ems_x++ ] = (((( row / (64/gsf)) * (8*gsf)) 
			   + (col / (64/gsf)) + 1 ) * (256/(gsf*gsf))) - 1;
   	}
   }
   sel_misc->image_size = 524288L;

   return ( p );     
}

/* Write image from EMS memory to hard disk. */

pv_ptr wthd ( register pv_ptr p )
{
   unsigned far *ems_data;               /* pointer to ems_data page */
   unsigned i, page, sequence, seconds;
   unsigned amtwritten;
   int fd;
   long fpos;
   char fname[13],  subdir[80], fullpathname[PATHSIZE];
   
     /* File header (fixed size 512 bytes) for file. */ 
 
     union file_hdr_union 
     {
       struct cam_structs_tag 
       { 	
         cam_misc misc;
         cam_sys  sys;
         cam_exp  exp;
       } structs;
       char     filler[512];
     } file_hdr;

     /* Calculate size of data to write based on current settings. */
      
     unsigned long size =   (unsigned long) sel_exp->cols 
                        * (unsigned long) sel_exp->rows 
                        / ( (unsigned long) sel_exp->sbin 
                          * (unsigned long) sel_exp->pbin );

     if ( ( sel_misc->image_size == 0L ) || ( sel_misc->ems_handle == 0 ) )
     {
	sprintf( omsg, "No image available to write for this camera.\n" );
	disp_omsg();
	return ( p );
     }

     /* Build sub-directory name based on exposure type, 
        and update sequence number. */
     
     switch ( sel_misc->exposure_type )
     {
        case 'A':
        	sprintf( subdir, "AURORAL\\CAM%d\\FILT%d\\", 
        		sel_cam, sel_misc->fw );
        	sequence = ki_data.auroral[ sel_cam ][ sel_misc->fw ]++;
     		break;
        case 'B':
        	sprintf( subdir, "BIAS\\CAM%d\\", sel_cam );
        	sequence = ki_data.bias[ sel_cam ]++;
     		break;
        case 'C':
        	sprintf( subdir, "CALIB\\CAM%d\\FILT%d\\", 
        		sel_cam, sel_misc->fw ); 
        	sequence = ki_data.calib[ sel_cam][ sel_misc->fw ]++;
     		break;
        case 'D':
        	sprintf( subdir, "DARK\\CAM%d\\", sel_cam ); 
        	sequence = ki_data.dark[ sel_cam ]++;
     		break;
        case 'F':
        	sprintf( subdir, "FLAT\\CAM%d\\FILT%d\\", 
        		sel_cam, sel_misc->fw ); 
        	sequence = ki_data.flat[ sel_cam ][ sel_misc->fw ]++;
     		break;
        case 'G':
        	sprintf( subdir, "AIRGLOW\\CAM%d\\FILT%d\\", 
        		sel_cam, sel_misc->fw );
        	sequence =  ki_data.airglow[ sel_cam ][ sel_misc->fw ]++;
     		break;
     }
     
     if ( sel_misc->exp_scale == 1 )				/* ms */
        seconds = (unsigned int) (sel_exp->exposure / 1000);
     else if ( sel_misc->exp_scale == 1000 )			/* Seconds */
        seconds = (unsigned int) sel_exp->exposure;
     else if ( sel_misc->exp_scale == 60000 )			/* Minutes */
        seconds = (unsigned int) (sel_exp->exposure * 60);
     if ( seconds > 999 )
        seconds = 999;

     /* Filename is "TDDCFSSS.NNN" where:
      * 
      * T   = Image type Airglow/Aurora, Bias, Calibration, Dark, Flat-field.
      * DD  = Day of month.
      * C   = Camera ( 0 or 1 ) 
      * F   = Filter #
      * SSS = Seconds of exposure.  (integral)
      * NNN = Unique sequence #.
      */
     
     /* Build file name based on current data. */
   
     sprintf( fname, "%c%2d%d%d%3d.%3d",
            sel_misc->exposure_type,
            sel_misc->dt.day,
            sel_cam,
            sel_misc->fw,
            seconds,
            sequence );

     for ( i=0; i<13; i++ )	/* change spaces to zeros in filename. */
       if ( fname[i] == ' ' )
          fname[i] = '0';

     strcpy( fullpathname, imagepath );
     strcat( fullpathname, subdir );
     strcat( fullpathname, fname );

     /* Reset pixel min and max. */

     sel_misc->pixel_min = 32767;
     sel_misc->pixel_max = 0;

     /* Returns pointer to EMS page frame. */

     ems_data = (unsigned *) ems_get_page_frame(); 
     if( !ems_data ) 
     {
	sprintf( omsg, "Unable to locate EMS page frame. Unable to proceed.\n");
	disp_omsg();
	return ( p );
     }
printf("%s\n", fullpathname);
     fd = open( fullpathname, O_WRONLY | O_CREAT |O_BINARY , 
                              S_IREAD | S_IWRITE );
     if ( fd < 0 )
     {
     	sprintf ( omsg, "Unable to open output file %s. (disk full ?)\n",
     			fullpathname );
     	disp_omsg();
        return ( p );
     }
     
     /* Write the image data to file starting at byte offset 512 */
     
     fpos = lseek( fd, 512L, SEEK_SET );
     if ( fpos != 512 )
     {
     	   sprintf ( omsg, "Unable to position file data pointer.\n" );
     	   disp_omsg();
           close(fd);
           return ( p );
     }
     

     for( page = 0; 
          page < (unsigned) ( 32 / ( sel_exp->sbin * sel_exp->pbin ) ); 
          page++ ) 
     {
        ems_map_page( 0, sel_misc->ems_handle, page );
        amtwritten = write( fd, ems_data, 16384 );    /* write one EMS memory page */
        if ( amtwritten != 16384 )
        {
     	   sprintf ( omsg, "Unable to write file data (disk full ?)\n" );
     	   disp_omsg();
           close(fd);
           return ( p );
        }
        scan_min_max( ems_data );
     }
     
     memset ( &file_hdr, ' ', 512 );	/* Clear file header */

     /* Copy camera structures to file header, then write it out. */

     memcpy ( &file_hdr.structs.misc, sel_misc, sizeof( cam_misc ) );
     memcpy ( &file_hdr.structs.sys,  sel_sys,  sizeof( cam_sys ) );
     memcpy ( &file_hdr.structs.exp,  sel_exp,  sizeof( cam_exp ) );

     /* Write the file header data to file starting at byte offset zero. */
     
     fpos = lseek( fd, 0L, SEEK_SET );
     if ( fpos != 0 )
     {
     	   sprintf ( omsg, "Unable to position file data pointer.\n" );
     	   disp_omsg();
           close(fd);
           return ( p );
     }
     
     amtwritten = write( fd, &file_hdr,  512 );
     if ( amtwritten != 512 )
     {
     	sprintf ( omsg, "Unable to write file header (disk full ?)\n" );
     	disp_omsg();
        close(fd);
        return ( p );
     }

   close(fd);

   /* Write updated file sequence number, and file name of last image. */

   if ( sel_cam == 0 )
   	strcpy ( ki_data.last_cam0_img_flnm, fullpathname );
   else
   	strcpy ( ki_data.last_cam1_img_flnm, fullpathname );
   	
   lseek( fdata, 0L, SEEK_SET );
   write( fdata, &ki_data, sizeof( KI_DATA ) );

   return ( p );
}

/* Scan current EMS page for find new MIN and MAX pixel intensities.
 * Optimized for speed.
 */

void scan_min_max ( unsigned far *page )
{
	register int 	min, max;
	register unsigned i, val;
	
	min = sel_misc->pixel_min;	/* local for speed. */
	max = sel_misc->pixel_max;
	for ( i=0; i<8192; i++ )
	{
		val = page[ i ];
		if ( val < min )
			min = val;
		if ( val > max )
			max = val;
	}
	sel_misc->pixel_min = min;
	sel_misc->pixel_max = max;
}

/* Set Operating mode - 1 or 2 camera operation. */

pv_ptr op_mode ( register pv_ptr p )	
{
     unsigned typ, opmode, x;
 
     /* Set up for one or two camera operation, one or two AFB operation. 
      *
      * Parameter = 0, 1, or 2 for Camera 0 only, camera 1 only, or both.
      */ 

     p = clause ( p, &typ, &opmode );	/* Get Camera op_mode value */
     if ( typ != CONST )
     {
     	sprintf( omsg,  "ERROR: 'Camera opmode' value expected.\n" );
     	disp_omsg();
     	sprintf( omsg,  "value = 0, 1, or 2 to enable cam0, cam1, or both\n" );
     	disp_omsg();
     	return ( p );
     }     
     if (( opmode != 0 ) && ( opmode != 1 ) && ( opmode != 2 ) )
     {
     	sprintf( omsg,  "ERROR: Camera opmode value should be 0, 1, or 2.\n");
     	disp_omsg();
     	return ( p );
     }
     
     /* If single camera mode is selected, make selected camera current. */ 
   
     if ( opmode != 2 )
     	select_camera( opmode );
     	
     /* Value stored is 1, 2, or 3 to allow bitwise testing. */
    
     std0_misc.cam_opmode = ++opmode;	/* Show whether 1 or 2 camera */
     std1_misc.cam_opmode = opmode;	/* operation is in effect. */
     
     return ( p );
}



