/******************************************************************************
 DECIDE.C

 Decision analysis program.  
 
 This program reads through the DECISION.DAT file produced by COMPARE.C 
 and counts the number of "DECISION 1" and "DECISION 0" results.  If 
 there are more "DECISION 1" results than "DECISION 0" results, a true value
 (1) is returned, otherwise a false (0) value is returned.  Any exception
 results in a -1 return.  The return value is used by an MS-DOS batch file
 to determine which observation mode will be executed next. 
 
******************************************************************************/
#include <bios.h>
#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#include "at200.h"
#include "idecls.h"
#include "camsys.h"

/* Function prototypes. */

void get_fname( void );
void get_lookup( void );
void show_help( void );
void scan_image( void );
void show_hdr( void );

/* Globals */

char filespath[PATHSIZE], fname[80], lkupflnm[80];
int fd;

void show_help(void)
{

  printf("USAGE: DECIDE [-d]\n"
         "  Where:\n"
         "    -d            Display results\n" );
  exit( -1 );
}

void main(int argc,char *argv[])
{
unsigned int ii, jj, x, decision, dispsw;
FILE *fp;
char	databuff[82], *env;

	/* Process command line arguments. */

	if ( ( env = getenv( "KI_FILES" ) ) != (char *) NULL )
	{
		strcpy( filespath, env );	/* Base path for KI files. */
		strcat( filespath, "\\" );
	}
	else
		filespath[0] = '\0';

	for( x=1 ; x < (unsigned) argc ; x++ ) 
	{
    	    if( argv[x][0] == '-' )
    	    {
    		switch( argv[x][1] ) 
    		{
      	 
		/* Display output switch. */

      		case 'd': case 'D':
      	  		dispsw = 1;
      	  		break;
      	 
        	default:
			show_help();
        		break;
		}
	    }
	    else
		show_help();
	}
	
	strcpy ( databuff, filespath );
	strcat ( databuff, "DECISION.DAT" );

	if ( ( fp=fopen( databuff, "r" ) ) == NULL ) /* Open for reading */
	{
		fprintf(stderr,"Can't open %s.\n", databuff );
		exit( -1 );
	}

	/* Read file, count occurences of "DECISION 1", 
	 * see if it is majority. 
	 */
	 
	ii = 0;				/* # of DECISION n lines in file */
	jj = 0;				/* # of DECISION 1 values in file */
	while ( fgets( databuff, 82, fp ) != (char *) NULL ) 
	{
		if ( !strncmp( &databuff[31], "DECISION", 8 ) )
		{
			ii++;
			if ( databuff[40] == '1' )
				jj++;
		}
	}

	/* Set decision again based on majority of "DECISION 1" 
	 * values over "DECISION 0" values.
	 */
	 
	if ( jj > ( ii / 2 ) )
		decision = 1;
	else
		decision = 0;
		
	if ( dispsw )
		printf( "Majority decision: %d out of %d results in return = %d\n",
			jj, ii, decision );

	/* Return with TRUE/FALSE majority decision. */

	exit( decision );
}
