/*  Keyboard driver for interpreter. Takes input one line at a time
 *  from stdin , generates a program vector and executes it .Loops
 *  until EOF. Note: Due to an apparent bug in the Venix i/o driver,
 *  EOF (ctrl-d) is ignored unless it is the first character in
 *  a line 
 */

#include <dos.h>
#include <errno.h>
#include <conio.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <signal.h>
#include <at200.h>
#include <camera.h>

#include "idecls.h"
#include "proto.h"
#define  MAIN_PROG	/* Needed for camsys.h */
#include "camsys.h"

/* Globals to this file. */

FILE 	*cmndlogfl;
int	cmndlogmode, templogmode, templogfl, dispmode;
int 	inhibit0, inhibit1, inhibit_display;
int	lastlogmin, lastlogsec, loginterval;
int	lastsamplemin, lastsamplesec, sampleinterval;
char 	templogflname[60];
char 	cmndlogflname[60];

char line[ MAXLINE+1 ];		/* KI command line. */
struct	diskfree_t df;
pv_ptr p = (void *) NULL;

/* Function prototypes used in this program. */

void _cdecl main ( int, char *[] );
int sabrti ( void );
int sabrtos ( void );
void init_ki ( void );
void init_struct ( void );
void sample_temperatures( void );
void show_usage( void );
void do_startup ( char * );
void terminate_ki ( void );

/* Display usage of KI */

void show_usage( void )
{
  printf("\nUsage: KI [-Start filename1] [-Xeq filename2]\n\n"
  	 "  where: '-Start filename1	Name of KI start/config file.\n"
	 "         '-Xeq   filename2    Name of script file to execute.\n" );
  exit ( 1 );
}

void _cdecl main( int argc, char *argv[] )
{
	register int len;
	register char *ep;
	int i, next;
	entry_ptr pv[50];
	char	startfile[PATHSIZE], startflname[50];
	
	_dos_getdate( &dt );
	std0_misc.dt = dt;
	std1_misc.dt = dt;
	
	_dos_gettime( &tm );
	std0_misc.tm = tm;
	std1_misc.tm = tm;
	
	abrt = NORMAL;
	signal(SIGINT, sabrti);
	signal(SIGQUIT, sabrtos);
	init_ki();
	st_init();

	lastlogmin = tm.minute;
	lastlogsec = tm.second;
	lastsamplemin = tm.minute;
	lastsamplesec = tm.second;
	
	/* Process command line args */
	
	cmndlogmode = templogmode = dispmode = 0;
	inhibit0 = inhibit1 = inhibit_display = 0;
	sampleinterval = 15;
	loginterval    = 120;
	line[0] = startflname[0] = '\0';
	
	for ( i=1; i<argc; i++ )
	{
		if ( argv[i][0] == '-' )
			switch( argv[i][1] )
			{
				
			/* Get startup file name. */ 	
			
			case 's':
			case 'S':
				   if ( ++i > argc )
				   	show_usage();
				   strcpy ( startflname, argv[ i ] );
				   break;

			case 'x':			/* XEQ <filename> */
			case 'X':
				   if ( ++i > argc )
				   	show_usage();
				   sprintf ( line, "xeq %s", argv[ i ] );
				   break;
				   
			default:   show_usage();
			}
		else
		{
			show_usage();
		}
	}
	
	/* Define full path to startup file. */

	if ( !startflname[0] )
	{
		strcpy ( startfile, filespath );
		strcat ( startfile, "KI_START.DAT" );
	}
	else
	{
		strcpy ( startfile, filespath );
		strcat ( startfile, startflname );
	}
	do_startup( startfile );	/* Process startup file. */

	/* Initialize AFB communications */
		
	com_init( std0_misc.portno );
	com_init( std1_misc.portno );

	/* Reset AFBs */

	select_camera( 1 );
	p = reset_afb( p );			/* Reset the AFB */
	select_camera( 0 );
	p = reset_afb( p );			/* Reset the AFB */

	if ( cmndlogmode )
	{
		fprintf( cmndlogfl, "KI started: %2d/%2d/%2d  %2d:%2d\n", 
			dt.year, dt.month, dt.day, tm.hour, tm.minute );
	}
	
	/* Main processing loop. Accept commands from keyboard or 
	 * script file until terminated by stop command, or program
	 * exception. 
	 */

	while (abrt != OSABORT) 
	{
		abrt = NORMAL;
		delete();		/* Clear symbol table. */
		printf ("*");
		if ( line[0] )		/* Initial command line args */
		{
			len = strlen ( line );
		}
		else if ( ( len = getline( stdin, line, MAXLINE ) ) == EOF) 
		{
			terminate_ki();
			exit(0);
		}
		if ( len == MAXLINE ) 
		{
			printf ("Input line too long.\n");
			continue; 
		}
		if ( len > 0 )
		{
			if ( ( ep = iparse( line, pv, &next ) ) != NULL ) 
				printf("%s",ep) ;
			else
			{
				pv[next] = NULL ;
				iexec(pv) ;
			}
		}
		line[0] = '\0';
	}
	
	terminate_ki();

}

/* Do normal termination things. */

void terminate_ki ( void )
{
	/* Program termination. */

	if ( cmndlogmode )
	{
		_dos_getdate( &dt );
		_dos_gettime( &tm );
		fprintf( cmndlogfl, "KI ended: %2d/%2d/%2d  %2d:%2d\n", 
			dt.year, dt.month, dt.day, tm.hour, tm.minute );
		fclose( cmndlogfl );
	}
	if ( templogmode )
		close( templogfl );

	close( fdata );		/* Close file sequence data. */

	/* Release EMS memory. */

        ems_free( std0_misc.ems_handle );
        ems_free( std1_misc.ems_handle );
}

/* set abrt to IABORT on interrupt signal */

int sabrti ( void )
{
	abrt = IABORT; 
	signal (SIGINT, sabrti) ;
	return(0) ;
}

/* set abrt to OSABORT on quit signal */

int sabrtos ( void )
{
	abrt = OSABORT ;
	signal(SIGQUIT, sabrtos);
	return(0);
}

/*  Get a line from file pointer fp and put in s. Returns length
 *  of line or EOF. 
 *  If input stream is 'stdin', then yeild control to a chron
 *  function to log temperatures at specified intervals.
 */

int getline ( FILE *fp, char *s, register int lim )
{
	register int c;
	int i, comment;
	unsigned int x = 0;
	
	i = comment = 0;
	
	while ( lim > 0 )
	{
	    c = 0;
	    if ( fp != stdin )
	    { 
	        /* Input from script file. */ 
	   
		if ( ( c = getc( fp ) ) == EOF ) 
			return (EOF);
	    }
	    else
	    {
	        /* Input stream is stdin. (console) */
	   
	    	if ( kbhit() )
	    	{
			if ( ( c = getche() ) == 0x1A ) 
				return (EOF);
	    	}
	    	else
	    	{
	    		/* No keyboard input, see if background needs time. */
	    		
	  		x++;
	    		if ( x == 0x2000 )
	    		{
	    			x = 0;
				chron_func();
	    		}
	    	}
	    }
	    if ( c )		/* got a character */
	    {
		/* Ignore input lines with '!' to end of line. */

		if ( c == '!' )		/* Introduce comments with '!' char. */
			comment = i;    /* Store comment position */
		else if ( c == 0x08 )	/* backspace */
			if ( i )
			{
				printf( " \b" );
				s[ i-- ] = '\0';
				lim++;
		   	}

		if ( c != 0x08 )	/* add character to input string */
		{
			s[i++] = (char) c;
			lim--;
		}
		
		if ( c == '\r' )	/* C/R character same as newline. */
			c = '\n';

		if ( c == '\n' )
		{ 
			s[ --i ] = '\0' ;
			if ( cmndlogmode )
			{
				fprintf( cmndlogfl, "%s\n", s );
			}
			
			/* If input from stdin, send newline. */	
	
			if ( fp == stdin )
				printf( "\n" );
			else if ( dispmode )	/* Display script line */
			{
				printf( "%s\n", s );
			}
			if (comment )		/* Truncate comment portion */
				s[ comment ] = '\0';
			return ( i ) ;
		}
	    }
	}


	/*  Scan to end of line to prevent excess characters from
	 *  appearing as next line . 
	 */

	while (!((c = getchar()) == EOF || c == '\n'))
	;
	return ( i ) ;
}

void init_ki ( void )	/* Initialize camera structures. */
{
  char *env, fname[80];
  unsigned int count;
  
  /* Check for EMM capability. */
  
  if ( !emm_present() )
  {
  	printf( "EMM capability not present - check system configuration.\n" );
  	exit( -1 );
  } 
  
  if ( emm_version() < 0x40 )
  {
  	printf( "EMM version incorrect - check system configuration.\n" );
  	exit( -1 );
  } 
  
  ems_data = (unsigned *) ems_get_page_frame();
  if ( !ems_data )
  {
  	printf( "Unable to locate EMS page frame - check system configuration.\n" );
  	exit( -1 );
  }
  
  std0_misc.cam_opmode = 3;		/* Both cameras enabled. */
  std1_misc.cam_opmode = 3;
  
  /* Initialize camera 1 structures. */
 
  select_camera( 1 );
  init_struct();

  /* then camera 0 */

  select_camera( 0 );
  init_struct();

  /* Get environment variables for path names. */

  if ( ( env = getenv( "KI_LOG" ) ) != (char *) NULL )
  {
     strcpy( templogflname, env );	/* Build temperature log file path. */
     strcat( templogflname, "\\" );
     strcpy( cmndlogflname, env );	/* Build command log file path. */
     strcat( cmndlogflname, "\\" );
  }
  else
  {
     templogflname[0] = '\0';		/* No log file path specified. */
     cmndlogflname[0] = '\0';
  }

  if ( ( env = getenv( "KI_IMAGE" ) ) != (char *) NULL )
  {
     strcpy( imagepath, env );		/* Base path for images. */
     strcat( imagepath, "\\" );
  }
  else
     imagepath[0] = '\0';

  if ( ( env = getenv( "KI_SCRIPT" ) ) != (char *) NULL )
  {
     strcpy( scriptpath, env );		/* Base path for XEQ script files. */
     strcat( scriptpath, "\\" );
  }
  else
     scriptpath[0] = '\0';

  if ( ( env = getenv( "KI_FILES" ) ) != (char *) NULL )
  {
     strcpy( filespath, env );		/* Base path for images. */
     strcat( filespath, "\\" );
  }
  else
     filespath[0] = '\0';

  /* Open KI sequence data file. */

  strcpy ( fname, filespath );
  strcat ( fname, "KI_DATA.DAT" );
  fdata = open( fname, O_RDWR | O_CREAT | O_BINARY , S_IREAD | S_IWRITE );
  if ( fdata < 0 )
  {
     	sprintf ( omsg, "Unable to open file %s. (disk full ?)\n", fname );
     	disp_omsg();
     	exit( -1 );
  }

  /* Get the file sequence numbers. */

  count = read ( fdata, &ki_data, sizeof ( KI_DATA ) );
  if ( ( count < 1 ) || ( ki_data.dt.day != dt.day ) )	/* Empty file ? */
  {
     memset( &ki_data, '\0', sizeof ( KI_DATA ) );	/* reset to zeros */
     ki_data.dt = dt;					/* Set effective date */
  }
}

void init_struct ( void )
{
  unsigned int i, j, x;
  
  /* Allocate EMS memory for this structure. */   
     
  i = ems_alloc( 524288L );	/* Maximum image size 512X512 X 16 bits. */

  if( !i ) 
  {
	printf( "Unable to allocate EMS memory. Cannot proceed.\n");
	exit( -1 );
  }

  sel_misc->ems_handle = i;
  sel_misc->image_size = 0L;

  sel_sys->ccd = ccd[4];	/* CCD type - TH7985B 512 X 512 pixels */
  if ( sel_cam == 0 )
  {
     sel_sys->at200_id = 2;	/* Controller ID and base addr */
     at200_base = 0x220;
  }
  else
  {
     sel_sys->at200_id = 3;
     at200_base = 0x230;
  }
  sel_sys->frame = 0;		/* This is a full frame CCD */
  sel_sys->open_delay = 5;	
  sel_sys->close_delay = 5;
  sel_sys->preamp_delay = 7;
  sel_sys->speed = A200KHZ;	/* 200K Hz Analog to digital convertor */
  sel_sys->data_bits = 14;	/* with 14 bit/pixel resolution */
      
  sel_exp->exposure = 200;	/* default exposure */
  sel_exp->gain = 0;		/* 0 = 1X, 1 = 4X gain */
  sel_exp->mpp_mode = 1;	/* allow/disallow MPP exposures */
  sel_exp->corg = 0;		/* sub-array col origin, row origin */
  sel_exp->rorg = 0;
  sel_exp->cols = 512;		/* subarray # of rows, cols */
  sel_exp->rows = 512;
  sel_exp->pbin = 2;		/* Column binning factor */
  sel_exp->sbin = 2;		/* Row binning factor */
  sel_exp->clr_cnt = 5;		/* # of times to clear CCD before exposure */
  sel_exp->row_shift = 0;

  sel_misc->camera = (unsigned char) sel_cam;	/* Camera number. (constant) */
  sel_misc->fw = 0;		/* Filter number. (supplied by AFB) */
  sel_misc->filter_desc[0] = '\0';	/* Filter description */
  sel_misc->preamp = 0;		/* Preamp on/off during exposure. */
  sel_misc->exp_scale = 1;	/* Exposure scale 1=ms, 1000 = sec, 60000=min */
  for ( x=0; x<5; x++ )	 	/* Temperature readings from AFB */
  {
     sel_misc->temp.tei[x] = 0;
     sel_misc->temp.tee[x] = 0;
     sel_misc->temp.td[x]  = 0;
     sel_misc->temp.tcc[x] = 0;
     sel_misc->temp.tcf[x] = 0;
     sel_misc->temp.tfw[x] = 0;
  }
  sel_misc->sequence = 0;		/* file sequence # */
  sel_misc->exposure_type = 'A';	/* Airglow/Aurora, Bias, Dark, Flat */
  sel_misc->version = (1 * 16) + 0;	/* file/program version = 1.0 */
  strcpy ( sel_misc->jpeg_qual, "100" );	/* JPEG quality */
  strcpy ( sel_misc->comp_fact, "1.0" );	/* Compression factor */
  strcpy ( sel_misc->os_version, "5.0" );	/* Operating System version */
  sel_misc->lens_desc[0] = '\0'; 
  sel_misc->filter_desc[0] = '\0'; 
  sel_misc->location[0] = '\0'; 
  sel_misc->azimuth[0] = '\0'; 
  sel_misc->operator[0] = '\0'; 
  
  dsp_command( DSP_HALT );		/* Stop any operation in progress. */

  /* Load setup program into camera */
    
  x = load_dsp_program( dsp_cam200k, size_of_dsp_cam200k );

  if( !x ) 
  {
    printf("ERROR: Unable to load program into AT200 at addr %X\n", at200_base );
    exit( 1 );
  }

  /* Wait for DSP to complete initialization */
  
  for( x=1; x; x++ )
    if( !(inp( at200_base+DSP_INTERRUPT_STATUS) & (HOST_FLAG_2 | HOST_FLAG_3)) )
      break;  
  if( !x ) 
  {  
    printf("AT200 at addr %X is unresponsive or missing.  Unable to proceed.\n",
           at200_base );
    exit( 2 );
  }

  /* Reset filter descriptions. */
  
  for ( i=0; i<2; i++ )
     for ( j=0; j<5; j++ )
        filter_desc[i][j][0] = '\0';
        
  /* Clear AFB data conversion tables. */

  memset( &afb_const[0], '\0', sizeof ( AFB_CONST ) );
  memset( &afb_const[1], '\0', sizeof ( AFB_CONST ) );
        
}

/* Display message to user, and write to log if enabled. */

void disp_omsg( void )
{
	/* Inhibit display prevents output to screen during
	 * temperature logging. 
	 */
	 
	if ( !inhibit_display )
		printf( "%s", omsg );
	if ( cmndlogmode )
		fprintf( cmndlogfl, "%s", omsg );
}

/* Carry out background temperature logging task at specified intervals. */

void chron_func( void )
{
	int	count, kb_per_cluster, lastmin;
	static  int	diskspace, diskfree;
	struct  dosdate_t sav_dt;
	struct  dostime_t sav_tm;
	unsigned long	second;
static	unsigned long	lastsecond = 0L, 
			nextsamplesecond = 0L, 
			nextlogsecond = 0L;

	_dos_getdate( &dt );
	if ( dt.day != ki_data.dt.day )		/* Pumpkin time! */
	{
	     memset( &ki_data, '\0', sizeof ( KI_DATA ) ); /* reset file seq */
             ki_data.dt = dt;				   /* Set new date */
        }
        lastmin = tm.minute;
	_dos_gettime( &tm );

	second = ((long) tm.hour * 3600L) + ((long) tm.minute * 60L) 
			+ (long) tm.second;

	/* Calculate available disk space every minute. */

	if ( tm.minute != (unsigned char) lastmin )
	{	
		/* Calculate total and available diskspace in MB on C: */	

		_dos_getdiskfree ( 3, &df );

		kb_per_cluster = df.sectors_per_cluster *
			 df.bytes_per_sector / 1024;
	
		diskspace = (int) ( 	(long) df.total_clusters * 
				(long) kb_per_cluster / 1024L );
		diskfree = (int) ( 	(long) df.avail_clusters * 
				(long) kb_per_cluster / 1024L );
	}
	
	/*  ***** CHECK HERE FOR DISK SPACE < THRESHOLD FOR TAPE BACKUP *** */


	/* Check for time of day clock wraparound. */

	if ( second < lastsecond )
	{
		lastsecond = second;
		nextlogsecond = 0L;
		nextsamplesecond = 0L;
		return;
	}
	lastsecond = second;
	
	if ( !templogmode )		/* Return here if not doing */
		return;			/* temperature logging. */
		
	/* See if temperature sample interval time has expired. */

	if ( second > nextsamplesecond )
	{
		nextsamplesecond = second + (long) sampleinterval;
		sample_temperatures();
	}
	
	/* See if write temperature log interval time has expired. */

	if ( second > nextlogsecond )
	{
		/* Time to record temperature log data. */
		
		nextlogsecond = second + (long) loginterval;

		sprintf( omsg, "time: %2d:%2d:%2d.%2d  ", 
			tm.hour, tm.minute, tm.second, tm.hsecond );
/*		disp_omsg();*/
		
		sprintf( omsg, "Diskspace: %4d MB, Available: %4d MB\n", 
			 diskspace, diskfree );
/*		disp_omsg();*/

		sav_tm = std0_misc.tm;	/* Save exposure date/time */
		sav_dt = std0_misc.dt;

		std0_misc.tm = tm;	/* Plug in current time. */
		std0_misc.dt = dt;
		
		/* Write the temperature data for both cameras. */
		
		count = write( templogfl, &std0_misc, sizeof ( std0_misc ) );
		if ( count != sizeof ( std0_misc ))
		{
		   sprintf( omsg, "ERROR: Unable to write temperature log\n"); 
		   disp_omsg();
		}

		std0_misc.tm = sav_tm;	/* Restore exposure date/time */
		std0_misc.dt = sav_dt;
		
		sav_tm = std1_misc.tm;	/* Save exposure date/time */
		sav_dt = std1_misc.dt;

		std1_misc.tm = tm;	/* Plug in current time. */
		std1_misc.dt = dt;
		
		count = write( templogfl, &std1_misc, sizeof ( std1_misc ) );
		if ( count != sizeof ( std1_misc ))
		{
		   sprintf( omsg, "ERROR: Unable to write temperature log\n"); 
		   disp_omsg();
		}
		
		std1_misc.tm = sav_tm;	/* Restore exposure date/time */
		std1_misc.dt = sav_dt;
		
		/* Reset temperature data */
		
		memset( &std0_misc.temp, '\0', sizeof( temp_t ) );
		memset( &std1_misc.temp, '\0', sizeof( temp_t ) );
	}
}


/* This procedure reads all temperatures for both camera units, 
 * saves them in the camera structures.
 */

void sample_temperatures( void )
{
	pv_ptr p = (void *) NULL; /* used for linkage to temperature functs */
	int save_camera;
	
	if ( std0_misc.cam_opmode & 0x1 )	/* If camera 0 enabled. */
	{
		save_camera = sel_cam;		/* Save current camera # */
		select_camera( 0 );		/* Read camera 0 temps. */
		inhibit_display = inhibit0;	/* Prevent temperature display */
		p = afb_id ( p );		/* Store AFB ID */
		p = tee ( p );			/* Log temperature readings */
		p = tei ( p );
		p = td  ( p );
		p = tcf ( p );
		p = tfw ( p );
		p = heater ( p );		/* Heater on/off status */
		p = fan_stat ( p );		/* Fan status */
		p = acpower ( p );
		p = plus5 ( p );
		p = plus15 ( p );
		p = minus15 ( p );
		select_camera( save_camera );	/* re-select current camera */
	}

	if ( std1_misc.cam_opmode & 0x2 )	/* If camera 1 enabled. */
	{
		save_camera = sel_cam;		/* Save current camera # */
		select_camera( 1 );		/* Read camera 1 temps. */
		inhibit_display = inhibit1;
		p = afb_id ( p );
		p = tee ( p );
		p = tei ( p );
		p = td  ( p );
		p = tcf ( p );
		p = tfw ( p );
		p = heater ( p );
		p = fan_stat ( p );
		p = acpower ( p );
		p = plus5 ( p );
		p = plus15 ( p );
		p = minus15 ( p );
		select_camera( save_camera );	/* re-select current camera */
	}
	inhibit_display = 0;		/* re-enable console display */
}


/* Process KI startup file. 
 *
 * Input is in the form: -keyword [keyvalue] [keyvalue]
 * where keywords may have zero or more keyvalues to use.
 */

void do_startup ( char *startfilename )
{
char	*token;
int	v1, v2, done, get_next = 1;


	if ( !set_cmdsrc ( startfilename, CMD_FILE ) )
	{
		printf( "ERROR: Unable to open startup file: %s\n",
			startfilename );
		terminate_ki();
		exit ( -1 );
	}
	
	while ( 1 )
	{
	    if ( get_next )
		token = get_nxt_tok();

	    if ( token == (void *) NULL )
	    	break;
	    else if ( token[0] == '-' )		/* Keyword */
	    {
		if ( strmatch( &token[1], "temperature_log" ) )
		{
			templogmode = 1;	/* Temperature logging */

			/* See if file name specified. */
				   
			token = get_nxt_tok();

			if( token[0] == '-')
			{
				strcat ( templogflname, "KI_TEMP.LOG" );
				get_next = 0;
			}
			else 
			{
				strcat ( templogflname, token );
				get_next = 1;
			}
			templogfl = open( templogflname, 
				O_WRONLY | O_CREAT | O_APPEND | O_BINARY,
				S_IREAD | S_IWRITE );
		}
		else if ( strmatch ( &token[1], "log" ) )
		{
			cmndlogmode = 1;	/* Log commands */
			token = get_nxt_tok();

			if( token[0] == '-')
			{
				strcat ( cmndlogflname, "KI_CMND.LOG" );
				get_next = 0;
			}
			else 
			{
				strcat ( cmndlogflname, token );
				get_next = 1;
			}
			cmndlogfl = fopen ( cmndlogflname, "w+" );
		}
		else if ( strmatch ( &token[1], "display" ) )
		{
			dispmode = 1;	/* Display cmnds */
			get_next = 1;	/* from script files */
		}
		else if ( strmatch ( &token[1], "interval" ) )
		{
			sampleinterval = get_nxt_number();
			loginterval = get_nxt_number();
			get_next = 1;
		}
		else if ( strmatch ( &token[1], "location" ) )
		{
			token = get_nxt_tok();
			if ( strlen ( token ) > 80 )
				token[80] = '\0';
			strcpy ( std0_misc.location,	/* Common to both */
				token );		/* cameras. */
			strcpy ( std1_misc.location,
				token );
		}

		/* Process '-camera n m' line where:
		 * n = camera number (0 or 1)
		 * m = AFB serial port address for camera ( 1 thru 4)
		 */ 

		else if ( strmatch ( &token[1], "camera" ) )
		{
			v1 = get_nxt_number();		/* Camera # 0 or 1 */
			if ( !( (v1 == 0) || (v1 == 1) ) )
			{
				printf( "ERROR: Camera number must be 0 or 1\n");
				terminate_ki();
				exit ( -1 );
			}

			select_camera( v1 );

			v2 = get_nxt_number();		/* AFB serial portno */
			if ( !( (v2 == 1) || (v2 == 2) || (v2 == 3) || 
				(v2 == 4) ) )
			{
				printf("ERROR: AFB port number must be 1 thru 4\n");
				terminate_ki();
				exit ( -1 );
			}
			sel_misc->portno = (unsigned char) v2;

			done = 0;
			while ( !done )
			{
				token = get_nxt_tok();
				if ( strmatch ( &token[1], "filter" ) )
				{
				    v2 = get_nxt_number();	/* Filter # */
				    if ((v2 >= 0) && (v2 < 5 ))
				    {
					token = get_nxt_tok();
					if ( strlen ( token ) > 5 )
						token[5] = '\0';
					strcpy ( &filter_desc[v1][v2][0],
						token );
				    }
				    else
				    {
					printf("ERROR: Filter number must be 0 thru 4\n");
					terminate_ki();
					exit ( -1 );
				    }
				}
				else if ( strmatch ( &token[1], "lens" ) )
				{
					token = get_nxt_tok();
					if ( strlen ( token ) > 5 )
						token[5] = '\0';
					strcpy ( sel_misc->lens_desc,
						token );
				}
				else if ( strmatch ( &token[1], "azimuth" ) )
				{
					token = get_nxt_tok();
					if ( strlen ( token ) > 10 )
						token[10] = '\0';
					strcpy ( sel_misc->azimuth,
						token );
				}
				else if ( strmatch ( &token[1], "operator" ) )
				{
					token = get_nxt_tok();
					if ( strlen ( token ) > 3 )
						token[3] = '\0';
					strcpy ( sel_misc->operator,
						token );
				}
				else
					done = 1;
			}
			
			get_next = 0;
		}
		
		/* Process -AFB x y line. 
		 * x = AFB number
 		 * y = AFB ID
 		 */
 		 
		else if ( strmatch ( &token[1], "afb" ) )
		{
			v1 = get_nxt_number();			/* AFB # */
			if ( !( (v1 == 0) || (v1 == 1) ) )
			{
				printf("ERROR: AFB number must be 0 or 1\n");
				terminate_ki();
				exit ( -1 );
			}

			afb_const[v1].id = get_nxt_number();	/* AFB ID */

			/* Process all AFB data conversion setups. */

			done = 0;
			while ( !done )
			{
				token = get_nxt_tok();
				if ( strmatch ( &token[1], "tei" ) )
				{
				    afb_const[v1].tei[0] = get_nxt_number();
				    afb_const[v1].tei[1] = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "tee" ) )
				{
				    afb_const[v1].tee[0] = get_nxt_number();
				    afb_const[v1].tee[1] = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "td" ) )
				{
				    afb_const[v1].td[0] = get_nxt_number();
				    afb_const[v1].td[1] = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "tcf" ) )
				{
				    afb_const[v1].tcf[0] = get_nxt_number();
				    afb_const[v1].tcf[1] = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "tfw" ) )
				{
				    afb_const[v1].tfw[0] = get_nxt_number();
				    afb_const[v1].tfw[1] = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "plus5" ) )
				{
				    afb_const[v1].plus5 = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "plus15" ) )
				{
				    afb_const[v1].plus15 = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "minus15" ) )
				{
				    afb_const[v1].minus15 = get_nxt_number();
				}
				else
					done = 1;
			}
			get_next = 0;
		}
		else if ( strmatch ( &token[1], "stop" ) )
		{
			token = get_nxt_tok();
			inhibit0 = atoi ( token );

			token = get_nxt_tok();
			inhibit1 = atoi ( token );
			get_next = 1;
		}
		else if ( token == (void *) NULL)
			break;
		else
		{
			printf ( "ERROR: Unknown keyword '%s' in startup file '%s'\n",
				token,  startfilename );
			terminate_ki();
			exit ( 1 );
		}
	    }
	    else
	    {
		printf ( "ERROR: Unknown keyword '%s' in startup file '%s'\n",
			token,  startfilename );
		terminate_ki();
		exit ( 1 );
	    }
	}
}

/* Compare first 'n' chars of in_token to target.
 *
 * Return true if they match, false otherwise. 
 */
  
int strmatch ( char *in_token, char *target )
{
register int i, j;

	j = strlen ( in_token );
	if ( j )
	{
		for ( i=0; i<j; i++ )
		{
			if ( !target[ i ] )
				return 0;   /* in_token longer than target */
			if ( !( target[ i ] == in_token[ i ] ) )
				return 0;   /* No compare */
		}
	}
	else
		return 0;		    /* NULL string */
		
	return 1;
}

/* Get next token from startup command file or command line.
 * Return pointer to stringified token, or NULL if EOF.
 */
 
char *get_nxt_tok( void )
{
static char	outstring[80];
int		i, stopper;

	if ( cmdsrc == CMD_LINE )
	{
		if(( cmdln[cmdlnx] == '\0' ) || ( cmdln[cmdlnx] == '\n' ))
		{
			cmdsrc = 0;
			return( (void *) NULL );
		}
	}
	else if ( cmdsrc == CMD_FILE )
	{
		/* Get a line of input from the startup file. */

		while ( cmdln[cmdlnx] == '\n' )
		{
			if ( fgets ( cmdln, 100, f1 ) == (void *) NULL )
			{
				fclose ( f1 );		/* EOF */
				cmdsrc = 0;
				return( (void *) NULL );
			}
			cmdlnx = 0;
			strcpy ( omsg, cmdln );
			disp_omsg();
		}
	}
	else
	{
		return( (void *) NULL );
	}
	
	/* Skip over white space to next token. */
	
	while ( 	(cmdln[cmdlnx] == ' ' ) 
		     || (cmdln[cmdlnx] == '\t' ) )
		cmdlnx++;

	/* Copy next token to output string. */

	i = 0;
	if ( cmdln[cmdlnx] == '"' )
	{
		stopper = '"';
		cmdlnx++;
	}
	else
		stopper = ' ';
		
	while ( 1 )
	{
		if (	(cmdln[cmdlnx] == (char) stopper ) 
		     || (cmdln[cmdlnx] == '\t' ) 
		     || (cmdln[cmdlnx] == '\n' ) )
		     break;
		else
			outstring[ i++ ] = cmdln[ cmdlnx++ ];
		if ( i == 80 )
			break;
	}

	outstring[ i ] = '\0';

	return strlwr ( outstring );
}


/* Get next token from startup file and return as numeric integer value. 
 * Set the source of command line args for above two functions.
 */

int get_nxt_number ( void )
{
char	*token;
int	i, j;

	token = get_nxt_tok();
	j = strlen ( token );
	for ( i=0; i<j; i++ )
	if ( !((token[i] == '-') || ((token[i] >= '0') && (token[i] <= '9' ))))
	{
		printf("ERROR: Numeric value expected.\n");
		terminate_ki();
		exit( -1 );
	}
	return ( atoi ( token ) );	
}

/* Set source of command args to file, or buffer. */

int set_cmdsrc ( char *source, int type )
{
	cmdlnx = 0;
	cmdsrc = 0;
	if ( type == CMD_FILE )
	{
		if ( ( f1 = fopen ( source, "r" ) ) == (FILE *) NULL )
		{
			return ( 0 );
		}
		cmdln[ cmdlnx ] = '\n';	/* to force read from file */
	}
	else if ( type == CMD_LINE )
	{
		strcpy ( cmdln, source );
	}
	else
	{
		return ( 0 );
	}
	cmdsrc = type;
	return ( 1 );		/* success */
}
