/******************************************************************************
 SCAN.C

 Scan specified image, tabulate pixel intensity distribution by gray-scale
 level.

******************************************************************************/
#include <bios.h>
#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#include "at200.h"
#include "idecls.h"
#include "camsys.h"


   /* File header (fixed size 512 bytes) for file. */ 
 
   union file_hdr_union 
   {
      struct cam_structs_tag 
      { 	
         cam_misc misc;
         cam_sys  sys;
         cam_exp  exp;
      } structs;
      char     filler[512];
   } file_hdr;
   
    int		pixel_max, pixel_min;
    unsigned	image_pbin, image_sbin, image_rows, image_cols, gs;

    char *fname;

void show_hdr ( void )
{
    printf ( "MISC fields\n" );
    printf (  "preamp      = %4d       ", file_hdr.structs.misc.preamp ); 
    printf (  "ems_handle  = %4d\n",      file_hdr.structs.misc.ems_handle );
    printf (  "exp_scale   =%5u       ",  file_hdr.structs.misc.exp_scale );
    printf (  "exp_type    =    %c\n",    file_hdr.structs.misc.exposure_type );
    printf (  "tee         = %4d       ", file_hdr.structs.misc.temp.tee[0] );
    printf (  "td          = %4d\n",      file_hdr.structs.misc.temp.td[0] );
    printf (  "tcc         = %4d       ", file_hdr.structs.misc.temp.tcc[0] );
    printf (  "tcf         = %4d\n",      file_hdr.structs.misc.temp.tcf[0] );
    printf (  "tfw         = %4d       ", file_hdr.structs.misc.temp.tfw[0] );
    printf (  "tei         = %4d\n",      file_hdr.structs.misc.temp.tei[0] );
    printf (  "heater      = %4d       ", file_hdr.structs.misc.heater );
    printf (  "acpower     = %4d\n",      file_hdr.structs.misc.acpower );
    printf (  "pixel_min   =%5d       ",  file_hdr.structs.misc.pixel_min );
    printf (  "pixel_max   =%5d\n\n",     file_hdr.structs.misc.pixel_max );

    printf (  "CAM_EXP fields          " );
    printf (  "CAM_SYS fields\n" );
    printf (  "exposure    =%5ld      ",  file_hdr.structs.exp.exposure );
    printf (  "at200_id    = %4d\n",      file_hdr.structs.sys.at200_id );
    printf (  "gain        = %4d      ",  file_hdr.structs.exp.gain );
    printf (  "frame       = %4d\n",      file_hdr.structs.sys.frame );
    printf (  "corg        = %4d      ",  file_hdr.structs.exp.corg );
    printf (  "open_delay  = %4d\n",      file_hdr.structs.sys.open_delay );
    printf (  "rorg        = %4d      ",  file_hdr.structs.exp.rorg );
    printf (  "close_delay = %4d\n",      file_hdr.structs.sys.close_delay );
    printf (  "cols        = %4d      ",  file_hdr.structs.exp.cols );
    printf (  "preamp_delay= %4d\n",      file_hdr.structs.sys.preamp_delay );
    printf (  "rows        = %4d      ",  file_hdr.structs.exp.rows );
    printf (  "speed       = %4d\n",      file_hdr.structs.sys.speed );
    printf (  "sbin        = %4d      ",  file_hdr.structs.exp.sbin );
    printf (  "data_bits   = %4d\n",      file_hdr.structs.sys.data_bits );

    printf (  "pbin        = %4d\n",      file_hdr.structs.exp.pbin );
    printf (  "clr_cnt     = %4d\n",      file_hdr.structs.exp.clr_cnt );
    printf (  "row_shift   = %4d\n\n",    file_hdr.structs.exp.row_shift );

    image_pbin = file_hdr.structs.exp.pbin;
    image_sbin = file_hdr.structs.exp.sbin;
    image_rows = file_hdr.structs.exp.rows;
    image_cols = file_hdr.structs.exp.cols;
    pixel_min  = file_hdr.structs.misc.pixel_min;
    pixel_max  = file_hdr.structs.misc.pixel_max;
}

 
void scan_image(void)
{
register unsigned row, col;
unsigned x, y, y2, y3, z, img_data[512];
unsigned long	count, hits[256];
int fd;

	/* Clear hits table */
	
	for ( x=0; x<256; x++ )
		hits[x] = 0L;

        /* Open image data file. */

	fd = open( fname, O_RDONLY | O_BINARY , S_IREAD | S_IWRITE );
	if ( !fd )
	{
		printf( "ERROR: Unable to open file: %s\n", fname );
		exit( 1 );
	}
	
	count = read( fd, &file_hdr, 512 );	/* Read file header */
	if ( count == 512 )
		show_hdr();
	else
	{
		printf( "ERROR: Unable to read header for file: %s\n", fname );
		exit( 1 );
	}
	
	x = image_rows / image_pbin;	/* x = actual # of pixels in x axis */
	y = image_cols / image_sbin;	/* y = actual # of pixels in y axis */
	
	count = 0L;

	/* Process the input image data */

	for( row=0; row<x; row++ )       /* Each row of image data */
	{
		read( fd, img_data, y*2 ); /* Read one row of data */
		
		for( col=0; col<y; col++ )    /* process this row */
		{
			if ( gs == 64 )
				z = ( img_data[col] >> 8 ) & 0x3F ;
			else
				z = ( img_data[col] >> 6 ) & 0xFF;
			hits[z] += 1;
			count++;
		}
	}
	
	close(fd);
	
	y = gs/4;
	y2 = y * 2;
	y3 = y * 3;
	printf ( "LVL     HITS\n" );
	printf ( "---    -----\n" );
	for ( x=0; x<y; x++ )
	{
	   printf( "%3d   %6ld    |    ", x,      hits[x]);
	   printf( "%3d   %6ld    |    ", x + y,  hits[x + y]);
	   printf( "%3d   %6ld    |    ", x + y2, hits[x + y2]);
	   printf( "%3d   %6ld \n",       x + y3, hits[x + y3]);
	}
}

void show_help(void)
{

  printf("USAGE: SCAN [-f filename] [-g 64 | 256]\n"
         "  Where:\n"
         "    -f `filename' is the name of the stored input image\n"
	 "    -g 64 | 256   Number of grey-scale levels to tabulate.\n");
  exit( 0 );
}

void main(int argc,char *argv[])
{
register unsigned long x;

 gs = 64;  
 if(argc < 3)
 show_help();

  for( x=1 ; x < (unsigned) argc ; x++ ) {
    if( argv[x][0] == '-' )
      switch( argv[x][1] ) {
        case 'f': case 'F':
          x++;
          fname=argv[x];
         break;
        case 'g': case 'G':
          x++;
	  gs = atoi(argv[x]);
	  if (( gs == 64 ) || ( gs == 256 ))
		break;
          else
          	show_help();
          break;
        default:
          show_help();
          break;
      }
    else
      show_help();
  }

  scan_image();
  exit( 0 );
}
