/******************************************************************************
 SHOW.C

 Displays image with 320X200 sub-array top left corner is given by command -q

 Displays the image with the serial register on the right edge of the screen.
 This program displays an image stored in a data file whose name is given on
 the command line. 

******************************************************************************/
#include <bios.h>
#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <graph.h>

#include <at200.h>

#include "idecls.h"
#include "camsys.h"


/* Globals */

KI_DATA ki_data;	/* Sequence numbers, and name of last image file */

   /* File header (fixed size 512 bytes) for file. */ 
 
   union file_hdr_union 
   {
      struct cam_structs_tag 
      { 	
         cam_misc misc;
         cam_sys  sys;
         cam_exp  exp;
      } structs;
      char     filler[512];
   } file_hdr;
   
    int		pixel_max, pixel_min;
    unsigned 	image_pbin, image_sbin, image_rows, image_cols, cam_no;

    char filespath[PATHSIZE], fname[80];
    unsigned xcoord=0;
    unsigned ycoord=0;

/*  screen_row[] is initialized in main() to point to the start of each row on
 *  the VGA display.  It is then used by the show_image() function to update
 *  the VGA memory. 
 */

char far *screen_row[200] = { 0, };   /* Pointers to graphics display rows */

/*  buffer[] is an array of 64 K pixels (128 K bytes) that is used to collect
 *  data from the camera.  The ptable[] array is used by the AT200 control
 *  library functions to collect the pixels.  It is initialized in main() to
 *  point to each 2 K pixel (4 K byte) block in the buffer[] array.  The ptr[]
 *  array is used by show_image() to display 320 pixel rows on the VGA.  It is
 *  also initialized in main(). 
 */

/* Display file header data. */

void show_hdr ( void )
{
    int x;
    
    printf ( "MISC fields\n" );
    printf (  "preamp      = %4d       ", file_hdr.structs.misc.preamp ); 
    printf (  "ems_handle  = %4d\n",      file_hdr.structs.misc.ems_handle );
    printf (  "exp_scale   =%5u       ",  file_hdr.structs.misc.exp_scale );
    printf (  "exp_type    =    %c\n",    file_hdr.structs.misc.exposure_type );
    printf (  "tee         = %4d       ", file_hdr.structs.misc.temp.tee[0] );
    printf (  "td          = %4d\n",      file_hdr.structs.misc.temp.td[0] );
    printf (  "tcc         = %4d       ", file_hdr.structs.misc.temp.tcc[0] );
    printf (  "tcf         = %4d\n",      file_hdr.structs.misc.temp.tcf[0] );
    printf (  "tfw         = %4d       ", file_hdr.structs.misc.temp.tfw[0] );
    printf (  "tei         = %4d\n",      file_hdr.structs.misc.temp.tei[0] );
    printf (  "heater      = %4d       ", file_hdr.structs.misc.heater );
    printf (  "acpower     = %4d\n",      file_hdr.structs.misc.acpower );
    printf (  "pixel_min   =%5d       ",  file_hdr.structs.misc.pixel_min );
    printf (  "pixel_max   =%5d\n\n",     file_hdr.structs.misc.pixel_max );

    printf (  "CAM_EXP fields          " );
    printf (  "CAM_SYS fields\n" );
    printf (  "exposure    =%5ld      ",  file_hdr.structs.exp.exposure );
    printf (  "at200_id    = %4d\n",      file_hdr.structs.sys.at200_id );
    printf (  "gain        = %4d      ",  file_hdr.structs.exp.gain );
    printf (  "frame       = %4d\n",      file_hdr.structs.sys.frame );
    printf (  "corg        = %4d      ",  file_hdr.structs.exp.corg );
    printf (  "open_delay  = %4d\n",      file_hdr.structs.sys.open_delay );
    printf (  "rorg        = %4d      ",  file_hdr.structs.exp.rorg );
    printf (  "close_delay = %4d\n",      file_hdr.structs.sys.close_delay );
    printf (  "cols        = %4d      ",  file_hdr.structs.exp.cols );
    printf (  "preamp_delay= %4d\n",      file_hdr.structs.sys.preamp_delay );
    printf (  "rows        = %4d      ",  file_hdr.structs.exp.rows );
    printf (  "speed       = %4d\n",      file_hdr.structs.sys.speed );
    printf (  "sbin        = %4d      ",  file_hdr.structs.exp.sbin );
    printf (  "data_bits   = %4d\n",      file_hdr.structs.sys.data_bits );

    printf (  "pbin        = %4d\n",      file_hdr.structs.exp.pbin );
    printf (  "clr_cnt     = %4d\n",      file_hdr.structs.exp.clr_cnt );
    printf (  "row_shift   = %4d\n\n",    file_hdr.structs.exp.row_shift );
    printf (" xcoord=%d,ycoord=%d\n",xcoord,ycoord);

    x = getch();

    image_pbin = file_hdr.structs.exp.pbin;
    image_sbin = file_hdr.structs.exp.sbin;
    image_rows = file_hdr.structs.exp.rows;
    image_cols = file_hdr.structs.exp.cols;
    pixel_min  = file_hdr.structs.misc.pixel_min;
    pixel_max  = file_hdr.structs.misc.pixel_max;

}

/* Select text mode or 320 x 200 - 256 color graphics mode */

void mcga_display_mode(int mode)
{ 
union REGS regs;

  regs.x.ax = (mode) ? 19 : 3;  /* AH=0, AL=Mode */
  int86(16, &regs, &regs);      /* Call video bios */
}

/* Set the RGB values for a color in 256 color graphics modes */

void mcga_color_set(int color,int red,int green,int blue)
{ 
  outp(0x3C8,color);  /* DAC color control register selects the color */
  outp(0x3C9,red);    /* Set Red DAC (0..63) */
  outp(0x3C9,green);  /* Set Green DAC (0..63) */
  outp(0x3C9,blue);   /* Set Blue DAC (0..63) */
}

void far *normalize_pointer(void far *data)
/* Make pointers have the lowest offset possible (and the largest segment).
   This function is needed to insure that each pointer in the ptable[] array
   can be used for 2 K pixels (4 K bytes).  Taking the address of an element
   of a huge array only guarantees that the segment:offset pair will be valid
   for that element.  The calculation performed here guarantees that the
   segment:offset pair will be valid for at least the 4 K addresses we need.
   See the Microsoft Macro Assembler 5.1 manual, section 13.2 for more
   information on REAL MODE ADDRESSING. */
{
unsigned seg,offs;

  seg  = (unsigned) ((long) data >> 16L);
  offs = (unsigned) ((long) data & 0xFFFF);
  seg += offs >> 4;
  offs &= 15;
  return( (void far *) (((long) seg << 16L) | (long) offs) );
}

/* Get name of last image file that was written. */

void get_fname( void )
{
int	fdata, count;
char	data_fname[ 80 ];

  /* Open KI sequence data file. */

  strcpy ( data_fname, filespath );
  strcat ( data_fname, "KI_DATA.DAT" );
  fdata = open( data_fname, O_RDWR | O_BINARY , S_IREAD | S_IWRITE );
  if ( fdata < 0 )
  {
     	printf ( "Unable to open file %s.\n", data_fname );
     	exit( -1 );
  }

  /* Read the data record (only one). */

  count = read ( fdata, &ki_data, sizeof ( KI_DATA ) );
  if ( count < 1 )					/* Empty file ? */
  {
	printf( "ERROR: file %s is empty.\n", data_fname );
	exit( -1 );
  }
  
  /* pick up the image file name depending on camera # */ 

  if ( cam_no == 0 )
  	strcpy ( fname, ki_data.last_cam0_img_flnm );
  else
  	strcpy ( fname, ki_data.last_cam1_img_flnm );
  close ( fdata );
}

/* Display an image from an image file.  
 * Data is read and processed in one pass without first being
 * all read into memory.
 */
 
void show_image(void)
{
unsigned char buff[320];
register unsigned row, col, scrn_x, scrn_y;
unsigned x, y, z, count, img_data[512];
int fd;

	/* Inititialize graphics display array */

	for( x=0 ; x<200 ; x++ )
		screen_row[x] = &( (char far *) 0xA0000000 )[x * 320];

        /* Open image data file. */

	fd = open( fname, O_RDONLY | O_BINARY , S_IREAD | S_IWRITE );
	if ( !fd )
	{
		printf( "ERROR: Unable to open file: %s\n", fname );
		exit( 1 );
	}
	
	count = read( fd, &file_hdr, 512 );	/* Read file header */
	if ( count == 512 )
		show_hdr();
	else
	{
		printf( "ERROR: Unable to read header for file: %s\n", fname );
		exit( 1 );
	}
	
	if ( _setvideomode( _MRES256COLOR ) == 0 )
	{
		printf ( "VGA mode required for this program.\n");
		exit( 0 );
	}

	/* Set color pallette to render 64 shades of grey */

	for( y=0; y<64; y++ )
		mcga_color_set(y,y,y,y);

	x = image_rows / image_pbin;	/* x = actual # of pixels in x axis */
	y = image_cols / image_sbin;	/* y = actual # of pixels in y axis */

        /* Calculate intensity expansion factor. 
         * Result is such that screen color values result in a range 0..63 
         * selecting 64 grey scale levels.
         */

	z = (( pixel_max - pixel_min + 62 ) / 63 );
	if( !z )
		z = 1;

	
	/* Process the input image data */

	for( row=0, scrn_y=0; row<x; row++ )       /* Each row of image data */
	{
		count = read( fd, img_data, y*2 ); /* Read one row of data */
		
		for( col=0, scrn_x=0; col<y; col++ )    /* process this row */
		{
		   if ( ( col >= xcoord ) && ( col < ( xcoord + 320 ) ) )
		      buff[scrn_x++] = (unsigned char) 
		      			((img_data[col] - pixel_min) / z );
		}
		if (scrn_x < 320 )			/* blank screen RHS */
			for ( ; scrn_x<320; scrn_x++ )
				buff[scrn_x] = '\0';
				
		if ( ( row >= ycoord ) && ( row < ( ycoord + 200 ) ) )
		   memcpy( screen_row[scrn_y++], buff, 320 );
	}
	if ( scrn_y < 200 )	/* Blank bottom of screen image */
	{
		for ( ; scrn_y<200; scrn_y++ )
		{
			for ( scrn_x=0; scrn_x<320; scrn_x++ )
				buff[scrn_x] = '\0';
				
			memcpy( screen_row[scrn_y], buff, 320 );
		}
	}

	close(fd);
	_settextposition( 25, 0 );
	_outtext ( fname );
	x = getch();			/* wait for user response */

	_setvideomode( _DEFAULTMODE );
}

void show_help(void)
{

  printf("USAGE: display [-f filename][-q # #][-c x]\n"
         "  Where:\n"
         "    -f `filename' is the name of the stored input image. (override)\n"
         "    -q #  #   # is the top left corner of the displayed sub-array,\n"
         "    -c x      Select camera 'x'. (default 0)\n" );
  exit( 0 );
}

void main(int argc,char *argv[])
{
register unsigned long x;
int data_bits = 14;
char	*env;
  
  fname[0] = '\0';
  cam_no = 0; 
 
  /* Process command line arguments. */

  if ( ( env = getenv( "KI_FILES" ) ) != (char *) NULL )
  {
	strcpy( filespath, env );	/* Base path for KI files. */
	strcat( filespath, "\\" );
  }
  else
	filespath[0] = '\0';

  for( x=1 ; x < (unsigned) argc ; x++ ) {
    if( argv[x][0] == '-' )
      switch( argv[x][1] ) {

        case 'c': case 'C':
          cam_no = atoi ( argv[ ++x ] );
      	  if (( cam_no != 0 ) && ( cam_no != 1 ))
      	  	show_help();
          break;
          
        case 'f': case 'F':
          x++;
          strcpy ( fname, argv[x] );
          break;

        case 'q': case 'Q':
          x++;
          xcoord = atoi(argv[x]);
          if( xcoord >512)
          show_help();
          x++;
          ycoord = atoi(argv[x]);
          if( ycoord >512)
          show_help();
          break;

        default:
          show_help();
          break;
      }
    else
      show_help();
  }

  /* Main line. */

  if ( !fname[0] )	/* Image file specified in command line? */
  	get_fname();	/* No - read it from KI_DATA file. */

  show_image();
  exit( 0 );
}
