#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <io.h>

#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>

#include <malloc.h>
#include <dos.h>
#include <graph.h>

#include "vga.h"
#include "pmtext.h"

#include "at200.h"
#include "idecls.h"
#include "proto.h"
#include "camsys.h"

/* Function prototypes */

extern void Select_Graphics(void);
extern void Select_Page( int );
extern void Write_Video( int, int, int, int, char far * );
extern void Draw_HLine( int, int, int, char );
extern void Draw_VLine( int, int, int, char );
void svga_color_set( int, int, int, int );
void Draw_colors( int, int, int, char );

void draw_frames( void );
void scan_image( unsigned char far * );
void print_string( int, int, char *, char );
void overlay_text( int );
void annotate ( int );

/* Globals */

unsigned int img_data[2][512];		/* Raw input image */
unsigned char pic0[256][256];		/* 256X256 by 64 level image */
unsigned char pic1[256][256];

unsigned char far *image0;
unsigned char far *image1;

union REGS regs;

pv_ptr view ( register pv_ptr p )	
{
int x, y, dx, dy, sav_camera;
int page = 0;
int color = 150;

	/* Returns pointer to EMS first page frame. */

	ems_data = (unsigned *) ems_get_page_frame(); 
	if( !ems_data ) 
	{
		sprintf( omsg, "Unable to locate EMS page frame. Unable to proceed.\n");
		disp_omsg();
		return p;
	}
   
	image0 = &pic0[0][0];
	image1 = &pic1[0][0];

	sav_camera = sel_cam;	/* Save selected camera number. */
	
	select_camera( 0 );
	scan_image( &pic0[0][0] );

	select_camera( 1 );
	scan_image( &pic1[0][0] );

	Select_Graphics();	/* Setup for SVGA display of images. */

	for (x = 0; x < 64; x++)
		svga_color_set(x, x, x, x);

	x = y = 0;
	dx = dy = 256;

	Write_Video( x, y, dx, dy, image0 );	/* Display first image. */

	x = 384;
	Write_Video( x, y, dx, dy, image1 );	/* Then second. */

	overlay_text( color );			/* Put in text. */
	draw_frames();				/* Frame the whole thing. */

	select_camera( 0 );
	annotate( color );

	select_camera( 1 );
	annotate( color );
	
	select_camera( sav_camera );		/* re-select current camera. */

	getch();				/* Wait for user response. */
	regs.x.ax = 0x0003;			/* Revert to normal video. */
	int86(0x10, &regs, &regs);

	return p;
}


void draw_frames()
{
char color = 163;

	Draw_HLine( 0, 639, 0, color );
	Draw_HLine( 0, 639, 255, color );
	Draw_HLine( 0, 639, 479, color );
	Draw_VLine( 0, 0, 479, color );
	Draw_VLine( 639, 0, 479, color );

	Draw_VLine( 255, 0, 255, color );
	Draw_VLine( 319, 0, 479, color );
	Draw_VLine( 383, 0, 255, color );
}

/* Read a camera image file, build 256 X 256 table of 8 bit pixel
 * intensity levels for display.  Accepts raw camera images of 512X512
 * (binning = 1) or 256X256 pixels (binning = 2).  
 */ 

void scan_image( unsigned char far *image )
{
register unsigned row, col;
unsigned int	ix, y, x, z, bin, count;
unsigned long	temp, ems_x;
unsigned int pixel_min, pixel_max, image_sbin, image_pbin;
unsigned int image_rows, image_cols;
unsigned int page, pos;

entry_ptr localpv[5];
int next;
char my_line[] = "compare \"-exit\"";	/*  will provide 'compare' args */
	
	/* Call 'comp' command to do image scan for min and max values,
	 * and histogram values.
	 */
	 
	iparse( my_line, &localpv[0], &next );
	localpv[ next ] = NULL;
	iexec( localpv );

	/* Get image characteristics. */
	
	image_pbin = sel_exp->pbin;
	image_sbin = sel_exp->sbin;
	image_rows = sel_exp->rows;
	image_cols = sel_exp->cols;

	/* Reset pixel min and max. */

	sel_misc->pixel_min = 32767;
	sel_misc->pixel_max = 0;

	/* Scan through EMS data once to get min and max pixel intensities. */

	for(	page = 0; 
        	page < (unsigned) ( 32 / ( image_sbin * image_pbin ) ); 
        	page++ ) 
	{
        	ems_map_page( 0, sel_misc->ems_handle, page );
		scan_min_max( ems_data );
	}
     
	pixel_min  = sel_misc->pixel_min;
	pixel_max  = sel_misc->pixel_max;


        /* Calculate intensity expansion factor. 
         * Result is such that screen color values result in a range 0..63 
         * selecting 64 grey scale levels.
         */

	z = (( pixel_max - pixel_min + 62 ) / 63 );
	if( !z )
		z = 1;

	ix = 0;

	/* Check if image to be processed is at an acceptable binning factor. */

	bin = 0;
	if ( ( image_pbin == 1 ) && ( image_sbin == 1 ) )
		bin = 1;
	else if ( ( image_pbin == 2 ) && ( image_sbin == 2 ) )
		bin = 2;
	if ( !bin )
	{
		printf("ERROR: binning must be 1X1, 2X2, for image.\n");
		exit ( -1 );
	}	      
	
	y = image_rows / image_pbin;	/* y = actual # of pixels in y axis */
	x = image_cols / image_sbin;	/* x = actual # of pixels in x axis */

	
	/* Process the input image data */

	ems_x = 0L;			/* index to int value in ems_data. */
	for( row=0; row<y; row++ )	/* Each row of image data */
	{
		/* 'pos' is the index to an word offset in ems_data.
		 * If zero, we are at the start of a new 16K (byte) page
		 */

		pos  = (unsigned int) ( ems_x % 8192L );
		if ( pos == 0 )
		{
			page = (unsigned int) ( ems_x / 8192L );
        		ems_map_page( 0, sel_misc->ems_handle, page );
		}
		
		/* Copy one row of camera data to img_data buffer. */	

	        memcpy ( &img_data[0][0], &ems_data[ pos ], x*2 );
	        ems_x += x;

		if ( bin == 1 )
		{
			/* If binning is one, we process two rows of
			 * camera data into one row of video pixels. 
			 */  	
	
			pos  = (unsigned int) ( ems_x % 8192L );
			if ( pos == 0 )
			{
				page = (unsigned int) ( ems_x / 8192L );
        			ems_map_page( 0, sel_misc->ems_handle, page );
			}
	        	memcpy ( &img_data[1][0], &ems_data[ pos ], x*2 );
	        	ems_x += x;
			row++;
		}

		/* Convert camera data to video image data. */
			
		for( temp = 0L, col=0; col<x; col++ )
		{
			/* If binning is one, average four pixels into one. */	
		
			if ( bin == 1 )
			{
				temp = img_data[0][col] + img_data[0][col+1]
					+img_data[1][col] + img_data[1][col+1];
				temp /= 4;
				col++;
			}
			else
				temp = img_data[0][col];
				
			/* Convert the 14 bit pixel intensity to modulo 64 */	
		
			image[ix++] = (unsigned char)((temp - pixel_min) / z);
		}
	}
}


/************************************************************************
*				pmtext.c				*
*		A module to display text labels on the display for	*
*	the Polar Camera.						*
************************************************************************/

void overlay_text( int color )
{
	/* First print all the labels that don't change */

	print_string( 1, 1, camera[ 0 ], (char) color );
	print_string( 384, 1, camera[ 1 ], (char) color );
	print_string( 256, 1, "16383", (char) color );
	print_string( 383 - 40, 1, "16383", (char) color );
	print_string( 256, 255 - 14, "0", (char) color );
	print_string( 375, 255 - 14, "0", (char) color );

	print_string( 1, 257, temperature[ CCD ], (char) color );
	print_string( 1, 257 + 40, temperature[ FW ], (char) color );
	print_string( 1, 257 + 80, temperature[ DOME ], (char) color );
	print_string( 1, 257 + 120, temperature[ CE ], (char) color );
	print_string( 1, 257 + 160, temperature[ EXT ], (char) color );

	print_string( 321, 257, temperature[ CCD ], (char) color );
	print_string( 321, 257 + 40, temperature[ FW ], (char) color );
	print_string( 321, 257 + 80, temperature[ DOME ], (char) color );
	print_string( 321, 257 + 120, temperature[ CE ], (char) color );
	print_string( 321, 257 + 160, temperature[ EXT ], (char) color );

	print_string( 1, 255 - 14, airglow, (char) color );
	print_string( 385, 255 - 14, airglow, (char) color );

	print_string( 1, 465, date_c, (char) color );
	print_string( 150, 465, time_c, (char) color );

}

void annotate( int color )
{
char	date_str[20], time_str[20];
int	i;
 
	sprintf( date_str, "%4d/%2d/%2d",
		sel_misc->dt.year,
		sel_misc->dt.month,
		sel_misc->dt.day );
	for ( i=0; i<11; i++ )
		if ( date_str[i] == ' ' )
			date_str[i] = '0';
	sprintf( time_str, "%2d:%2d:%2d",
		sel_misc->tm.hour,
		sel_misc->tm.minute,
		sel_misc->tm.second );
	for ( i=0; i<11; i++ )
		if ( time_str[i] == ' ' )
			time_str[i] = '0';

	print_string( 35, 465, date_str, (char) color );
	print_string( 185, 465, time_str, (char) color );
}

